#include "config.h"
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <unistd.h>
#include <errno.h>
#include "SnortLog.h"
#include "NetRangerLog.h"
#include "DragonLog.h"
#include "HttpUpload.h"

#define SFCLEAN_VERSTRING "1.01"
#define AUTO_RETRY 25


extern char *optarg;
int verbose = 0;
void usage()
{

  fprintf(stderr, "ARIS Extractor Version %s\n"
                  "Usage: extractor [options] -u <username> [-p password] <logfile>\n"
		  "       extractor [options] -a <username/password file> <logfile>\n"
		  "       extractor [options] -X <logfile>\n"
                  "Options:\n"
		  "  -u <username>\t\tUsername for upload server.\n"
	          "  -p <password>\t\tPassword for upload server.\n\t\t\t(ARIS Extractor will prompt for password if omitted.)\n"
		  "  -a <username/password file>\tText file containing Username and Password for upload server\n\t\t\t(Newline separated)\n"
	          "  -f <portscan_logfile>\tOptionally specify a Snort portscan logfile to process.\n"
	          "  -c <mask1,mask2...>\tStrip IP addresses in specified netmasks from output.\n"
	          "  -S <proxy_addr:port>\tUse a proxy server and optionally specify a port.\n"
                  "  -U <proxy_user>\tSpecify a username for authenticated proxy.\n"
                  "  -P <proxy_password>\tSpecify a password for authenticated proxy.\n"
	          "  -i <sensor_id>\tOptionally specify a unique sensor identifier.\n"
	          "  -X\t\t\tParse log only.  Do not upload results to server.\n"
	          "  -V\t\t\tDetailed output about log processing\n"
	          "  -h <host>\t\tOverride default upload server with <host>.\n"
		  "  -n\t\t\tParse Cisco Secure IDS (Netranger) Logfile.\n"
		  "  -d\t\t\tParse Dragon IDS Logfiles. <logfile> should specify the DB directory and driders.cfg file\n"
		      "\t\t\t(Comma separated)\n"
	          "\n  <logfile> is the path to an input logfile.\n\n", SFCLEAN_VERSTRING);
  fflush(stderr);
  exit(0);
}

char *homedir = NULL;


FILE *
openTimefile(char *sensor)
{
  char filename[1024];
  FILE *f;
  int n = 0;

  if(sensor)
    n = strlen(sensor);

  if(n > 100)
    return NULL;

  if(!homedir) {
    if(getenv("HOME"))
      homedir = strdup(getenv("HOME"));
    if( !homedir || !homedir[0]) {
      fprintf(stderr, "HOME environment variable must be set\n");
      fflush(stderr);
      exit(1);
    }
  }

  memset(filename, 0, sizeof(filename));
  snprintf(filename, 1023 - strlen("/lasttime-") - n, "%s/.sfclean", homedir);
  if(mkdir(filename, 0755) && (errno != EEXIST)) {
    fprintf(stderr, "Error creating directory: %s\n", filename);
    fflush(stderr);
    exit(1);
  }
  strcat(filename, "/lasttime");

  if(sensor) {
    strcat(filename, "-");
    strcat(filename, sensor);
  }

  if(!(f = fopen(filename, "r+")) && errno == ENOENT)
    return( fopen(filename, "w+") );
  else
    return f;
}


int 
loadTimestamp(double *ts, char *sensor){  
  
  FILE *f;
  char buf[100];

  if(!(f = openTimefile(sensor))) {
    fprintf(stderr, "Could not open timestamp file\n");
    fflush(stderr);
    return(-1);
  }
  
  buf[0] = 0;
  fgets(buf, 100, f);
  if(buf[0] == 0) { // file is empty
    *ts = 0.0;
  }
  else {
    if(sscanf(buf, "%lf", ts) != 1) {
      fprintf(stderr, "Error reading timestamp. Timstamp file contains '%s'\n", buf);
      fflush(stderr);
      return (-1);
    }
  }

  if(verbose)
    printf("Read saved timestamp = %.8f\n\n", *ts);

  return 0;
}
  
  
int
saveTimestamp(double ts, char *sensor)
{
  FILE *f;
  char buf[100];

  if(!(f = openTimefile(sensor))) {
    fprintf(stderr, "Could not open timestamp file\n");
    fflush(stderr);
    return(-1);
  }

  memset(buf, 0, sizeof(buf));
  snprintf(buf, 99, "%.8f", ts);
  fputs(buf, f);

  if(verbose)
    printf("Saved timestamp = %.8f\n\n", ts);
  return (0);
}


/* read username & password file */
int
readAuthFile (char *fname, char **username, char **password)
{
  char sAuth[32];
  char *newline = NULL;
  FILE *pAuthFile = fopen (fname, "r");
  if (!pAuthFile)
    return -1;
  if (!fgets (sAuth, 31, pAuthFile))
  {
    fclose (pAuthFile);
    return -1;
  }
  newline = strchr (sAuth, '\n');    
  if (newline)
    *newline = '\0';
  *username = strdup (sAuth);

  if (!fgets (sAuth, 31, pAuthFile))
  {
    free (*username);
    *username = NULL;
    fclose (pAuthFile);
    return -1;
  }
  newline = strchr (sAuth, '\n');    
  if (newline)
    *newline = '\0';
  *password = strdup (sAuth);

  fclose (pAuthFile);
  return 0;
}


int
main(int argc, char **argv)
{
  double last_date;
  double ts = 0.0;
  CIDSLog *myLog;
  HttpUpload *up;
  char op;
  char *pscanfile = NULL;
  int dopscan = 0;
  int parseonly = 0;
  int netranger=0;
  int dragonlog = 0;
  int authfile = 0;
  int authcmdline = 0;
  char *excludemasks = NULL;
  char *filenames = NULL;
  char *username = NULL;
  char *password = NULL;
  char *proxy_server = NULL;
  char *proxy_user = NULL;
  char *proxy_password = NULL;
  char *sensor_id = NULL;
  char *overridehost = NULL;
  char pathbuf[1024];
  char errbuf[256];
  struct stat stat_buf;
  int nRetry = AUTO_RETRY;
  bool bRetry = true;

  /* mask out SIGPIPEs */
  sigset_t set;
  sigemptyset (&set);
  sigaddset (&set, SIGPIPE);
  sigprocmask (SIG_BLOCK, &set, NULL);
  /* ================= */

  while((op = getopt(argc, argv, "ndXVf:c:u:p:i:S:U:P:h:a:")) != -1) 
    switch(op) {
    case 'd':
      dragonlog++;
      break;
    case 'n':
      netranger++;
      break;
    case 'h':
      overridehost=optarg;
      break;
    case 'X':
      parseonly ++;
      break;
    case 'V':
      verbose ++;
      break;
    case 'f':
      dopscan ++;
      pscanfile = optarg;
      break;
    case 'c':
      excludemasks = optarg;
      break;
    case 'a':
      if (authcmdline)			/* if the username or password are on the command line */
        break;				/* ignore the auth file */
      authfile ++;
      if (readAuthFile (optarg, &username, &password) == -1)
	authfile = -1;
      break;
    case 'u':
      authcmdline ++;
      if (authfile == 1) {              /* if the username & password are defined by the auth file */
        free (username);		/* override them */
        username = NULL;
        free (password);
        password = NULL;
      }
      authfile = 0;
      username = optarg;
      break;
    case 'p':
      authcmdline ++;
      if (authfile == 1) {		/* if the username & password are defined by the auth file */
	free (username);		/* override them */
	username = NULL;
	free (password);
	password = NULL;
      }
      authfile = 0;
      password = strdup(optarg);
      if(optarg)
       optarg[0] = 0; 
      break;
    case 'S':
      proxy_server = optarg;
      break;
    case 'U':
      proxy_user = optarg;
      break;
    case 'P':
      proxy_password = optarg;
      break;
    case 'i':
      sensor_id = optarg;
      break;
    case '?':
    default:
      usage();
    }

  if (authfile == -1) {
    fprintf(stderr, "Error reading username/password file\n\n");
    fflush(stderr);
    usage();
  }
  
  /* TODO: getpass () only returns 8 characters in Solaris (& possibly other unixes) */ 
  if(username && !password) 
#ifdef HAVE_GETPASSPHRASE
    password = strdup(getpassphrase("Enter Password: "));
#else
    password = strdup(getpass("Enter Password: "));
#endif

  if(!(username && password) && !parseonly)
    usage();

  if(proxy_server)
    if((proxy_user && !proxy_password) || (proxy_password && !proxy_user)) {
      fprintf(stderr, "Must specify both proxy username and password\n\n");
      fflush(stderr);
      usage();
    }

  if(!proxy_server && (proxy_password || proxy_user))
    usage();

  argc -= optind;
  argv += optind;
  if(argc != 1) 
    usage();

  if(!dragonlog && stat(argv[0], &stat_buf) == -1) {
    fprintf(stderr, "Could not open '%s'\n", argv[0]);
    fflush(stderr);
    exit(-1);
  }
  if(!dragonlog && S_ISDIR(stat_buf.st_mode)) {
    fprintf(stderr, "Input file '%s' is a directory.  It's supposed to be an IDS alert file\n", argv[0]);
    fflush(stderr);
    exit(-1);
  }
  
  if(dopscan) {
    if(netranger) {
      fprintf(stderr, "Portscan file option incompatible with Netranger option\n");
      fflush(stderr);
      exit(-1);
    }
    if(dragonlog) {
      fprintf(stderr, "Portscan file option incompatible with Dragon option\n");
      fflush(stderr);
      exit(-1);
    }
    if((filenames = (char *) malloc(strlen(argv[0]) + strlen(pscanfile) + 2)) == NULL) {
      fprintf(stderr, "Ack! No memory!\n");
      fflush(stderr);
      exit(-1);
    }
    strcpy(filenames, argv[0]);
    strcat(filenames, ",");
    strcat(filenames, pscanfile);
  }
  else 
    filenames = strdup(argv[0]);

  if(loadTimestamp(&ts, sensor_id))
    ts = 0.0;


  memset(pathbuf, 0, sizeof(pathbuf));
  if(netranger) {
    myLog = new CNetRangerLog;
    snprintf(pathbuf, sizeof(pathbuf) - 1, "%s/.sfclean/netranger.xml", homedir);
  }
  else if (dragonlog) {
    myLog = new CDragonLog;
    snprintf(pathbuf, sizeof(pathbuf) - 1, "%s/.sfclean/dragonlog.xml", homedir);
  }
  else {
    myLog = new CSnortLog;
    snprintf(pathbuf, sizeof(pathbuf) - 1, "%s/.sfclean/snortlog.xml", homedir);
  }

  if (excludemasks)
  { 
     if(!(myLog->SetIPMasks(excludemasks)))
     {
	 delete myLog;
         usage();
     }
  }
 
  if(!(myLog->Output(sensor_id, ts, 
		     filenames,
		     pathbuf, last_date))) {
    delete myLog;
    fprintf(stderr, "Error processing logfile\n");
    fflush(stderr);
    exit(1);
  }

  
  myLog->FreeIPMasks();
  if(filenames)
    free(filenames);

  // XXX add printStats() to CNetRangerLog and CDragonLog
  if(verbose && !netranger)
    ((CSnortLog *)myLog)->printStats();

  delete myLog;

  if(parseonly) {
    fprintf(stderr, "Exiting without submitting log to server\n"
	            "Output saved in %s\n", pathbuf);
    fflush(stderr);
    return 0;
  }

  bool bUploaded = false;
  while (bRetry)
  {
    up = new HttpUpload(pathbuf, username, password, "/tmp/sfclean.XXXXXX");

    if(proxy_server) 
      up->setProxyInfo(proxy_server, proxy_user, proxy_password);
    
    if(overridehost)
      up->setAlternateHost(overridehost);

    bUploaded = up->doTransfer(); 
    int errInfo = up->errorInfo(errbuf, 256);
    delete up;

	if ((errInfo == HttpUpload::ERR_TIMEOUT) || (errInfo == HttpUpload::ERR_WRITE) || (errInfo == HttpUpload::ERR_SERVER) || (errInfo == HttpUpload::ERR_SSL) || (errInfo == HttpUpload::ERR_SERVAUTH))
	{
		nRetry--;
		if (nRetry > 0)
		{	
			fprintf (stderr, "Retrying Upload...\n");
			continue;
		}
	}
	if (errInfo != HttpUpload::SUCCESS)
	{
	    fprintf(stderr, "\nError uploading log: %s\n", errbuf);
	    fflush(stderr);
	}
	bRetry = false;
  }

  if (bUploaded)
  {
	saveTimestamp(last_date, sensor_id);
	fprintf(stderr, "Upload completed\n");
	fflush(stderr);
  }
  return 0;
}
